"""This module contains all the methods to obtain information about the
model uncertainty.

"""


def rfunc_uncertainty(ml, rfunc, alpha):
    """

    Parameters
    ----------
    ml: pastas.Model
        Pastas model instance
    rfunc: pastas.rfunc.RfuncBase
        Pastas response function
    alpha: float
        float between 0 and 1 that indicates the

    Returns
    -------

    """

    return series
